/* readatq.c -- Copyright (C)  2002-2008 Marco Cesati
 * Thu, 09 Sep 2004 18:09:11 +0200
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#if HAVE_STDLIB_H
#    include <stdlib.h>
#endif

#include <stdio.h>

#include "../libatq/atqgrammar.h"

#define MAXGROUPS 10

int groups_s[MAXGROUPS] = { 0, };
int groups_t[MAXGROUPS] = { 0, };


#define panic(A...) do { fprintf(stderr,A); exit(EXIT_FAILURE); } while(0)



/*
    check_field() counts how many occurrences of every token are in
    the ATQ file. In particular, it increments a counter for the
    number of questions, whose address is passed as parameter. It also
    checks that no more than one occurrence exists for any field beside
    "QUESTION", "MULQUESTION", "GOOD", and "BAD". The function returns 1
    in case of error, 0 otherwise.
*/

int check_field(int ntoken, int *question)
{
    static int fcount[MAXTOKEN+1] = {0, };
    
    if (ntoken < 1 || ntoken > MAXTOKEN) {
        fprintf(stderr, "Unexpected token (%d)\n", ntoken);
        return 1;
    }
    
    switch (ntoken) {
    
        case QUESTION:
            ++*question;            
            return 0;

        case MULQUESTION:
            ++*question;
        /* FALL THROUGH */

        case GOOD:
        case BAD:
            return 0;

        default:
            if (!(fcount[ntoken]++)) /* first occurrence is fine */
                return 0;
                
            fprintf(stderr, "\nError in ATQ file: "
                            "%s field duplicated\n", token[ntoken]);
    }
    return 1;
}

/*
    format of the \group token: 
          <group-def> := <number>/<number>
          <group-line> := "\group{"<group-def>[[,<group-def>]...]"}"
    In any group definition, the first number represents the number
    of questions that will be selected among the group, and the second
    number represents the number of questions that build the group.
    The questions of a group are always sequential on the file, thus
    the groups are also defined by the ordering of the questions.
    The function fills the group_s[] and group_t[] arrays, and 
    initializes the counters ngroup, nquiz (the number of questions as seen
    by the student), and nquestions (the total number of questions
    in the file).
*/

int nquestion;

void count_groups(int *items, int numitem)
{
    int i, g;
    int nquiz = 0;
    
    nquestion = 0;

    if (numitem % 2)
        panic("\nError in ATQ file: Wrong format in groups\n"
              "(\\groups{a/b,c/d,e/f,...})\n");
    
    if (numitem > MAXGROUPS * 2)
        panic("\nError in ATQ file: File has too many groups (%d)\n"
              "(Maximum number of groups is %d)\n", numitem/2, MAXGROUPS);

    for (i=g=0; i<numitem; i++, g++) {

        groups_s[g] = items[i];
        groups_t[g] = items[++i];
           
            
        if (groups_s[g] > groups_t[g])
        
            panic("\nError in ATQ file: Group #%d (%d/%d) is not valid\n"
                  "(In a group definition, the second number cannot be smaller"
                  "\nthan the first number)\n", g, groups_s[g], groups_t[g]);

        if (!groups_s[g] || !groups_t[g])
            fprintf(stderr, "\nWARNING: value zero in group definition\n");

        nquiz += groups_s[g];
        nquestion += groups_t[g];        
    }
    printf("The ATQ file includes %d groups: (", numitem/2);
    for (i=0; i<numitem/2; ++i)
        printf(" %d/%d", groups_s[i], groups_t[i]);
    printf(" )\nEach test sheet includes %d questions, choosed among a grandtotal of %d\n",
           nquiz, nquestion);
}

inline int get_token(void)
{
    int t = yylex();
    if (t == ERR)
        panic(_errors[_error], "ATQ", _linenum);
    return t;
}

int main(void)
{
    int tok;
    int cquestion = 0;
 
    restart_atq_scanner(stdin);

    tok = get_token();
               
    while (tok) {

        switch(tok) {            
            case ARG:
            break;
            
            case NUM:
            break;
            
            case GROUPS:
            {
                int i = 0;
                int items[MAXGROUPS*2];

                check_field(tok, &cquestion);
                
                while ((tok = get_token())) {
                    if (tok != NUM)
                        break;
                    items[i++] = _value;
                }
                count_groups(items, i);
                continue;
            }
            
            default:
                if (check_field(tok, &cquestion))
                    panic("Aborted at line #%d\n", _linenum);
            break;
        }
        tok = get_token();
    }
    
    if (cquestion != nquestion)
        panic("Error in ATQ file: Wrong number of questions\n"
              "(In your \"%s\" field you specified a total number of "
              "%d questions,\nhowever there are %d occurrences of the \"%s\" "
              "and \"%s\" fields)\n", token[GROUPS], nquestion, cquestion, 
              token[QUESTION], token[MULQUESTION]);
              
    if (!nquestion)
        panic("\nCannot find any question in atq file!\n");

    return EXIT_SUCCESS;
}

