/* mktest.c -- Copyright (C)  2000-2008 Marco Cesati
 * Tue, 15 Jun 2004 19:59:36 +0200
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>

#if HAVE_STDLIB_H
#  include <stdlib.h>
#endif

#if HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include <signal.h>
#include <errno.h>
#include <time.h>

#if HAVE_STRING_H
#  include <string.h>
#endif

#if HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include "mktest.h"
#include "texfile.h"

#define VERSION_AND_COPYRIGHT \
          PACKAGE " version " PACKAGE_VERSION \
          " -- Copyright (C) 2000-2008 Marco Cesati <cesati@uniroma2.it>\n\n" \
          "This is free software; see the source for copying conditions.\n" \
          "There is NO warranty; not even for MERCHANTABILITY or FITNESS " \
          "FOR A PARTICULAR PURPOSE.\n\n"



void process_line(char const *);
static void create_form(int, int [], int);
void create_text(int, int [], int);
void scramble(int [], int);

char *formprogram = "mkform";

char const extension[] = ".atq";
char * basename;

static inline void _execute_child(const char *, char *const [],
                                  const char *, const char *)
                                  __attribute__((noreturn));

void _execute_child(const char *command, char *const argv[],
                           const char *redir_stdout, const char *redir_stderr)
{
    if (redir_stdout) {
        int fd = open(redir_stdout, O_CREAT|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR);
        if (fd == -1 || dup2(fd, 1) == -1 || close(fd) == -1) {
            perror(redir_stdout);
            exit(EXIT_FAILURE);
        }
    }
    if (redir_stderr) {
        int fd = open(redir_stderr, O_CREAT|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR);
        if (fd == -1 || dup2(fd, 2) == -1 || close(fd) == -1) {
            perror(redir_stderr);
            exit(EXIT_FAILURE);
        }
    }
    execvp(command, argv);
    fprintf(stderr, "Error (%d) while executing '%s'\n", errno, command);
    exit(EXIT_FAILURE);
} 

static int execute_command(char *command, char *const argv[],
                           char *redir_stdout, char *redir_stderr)
              
{
   pid_t child;
   int rc, status;

   child = fork();
   if (child == -1) {
       fprintf(stderr, "Cannot create child process\n");
       exit(EXIT_FAILURE);
   }

   if (!child)
       _execute_child(command, argv, redir_stdout, redir_stderr);

   rc = waitpid(child, &status, 0);
   return (rc == child && WIFEXITED(status) ? WEXITSTATUS(status) : -1);
}

static int error_in_latex_compilation(const char *text, const char *file)
{
        FILE *f;
        char line[80];
        int r;

        f = fopen(file, "r");
        if (!f)
            return 1;
        r = !fgets(line, 80, f);
        fclose(f);
        return r || strncmp(line, text, strlen(text));
}
static int min_scores[9*MAXQUESTIONS] = { 0, };
static int max_scores[9*MAXQUESTIONS] = { 0, };

static void update_form_stats(int quiz[], int nquiz)
{
   int i, totmin, totmax;
   totmin=totmax=0;
   for (i=0;i<nquiz;++i) {
      int min, max, j, q = quiz[i];
      int *p = point+q*MAXANSWERSPERQUESTION;
      min = max = *p++;
      for (j=1; j<nanswers[q]; ++j,++p) {
          if (*p < min)
              min = *p;
          if (*p > max)
              max = *p;
      }
      totmin += min;
      totmax += max;
   }
   min_scores[-totmin]++;
   max_scores[totmax]++;
}

static void show_form_stats(void)
{
    int i, *p = min_scores;
    fprintf(stderr, "\rMin/max scores of the forms: ");
    for (i=0; i<9*MAXQUESTIONS; ++i, ++p)
        if (*p > 0)
            fprintf(stderr, " %d(%d)", -i, *p);
    p = max_scores;
    for (i=0; i<9*MAXQUESTIONS; ++i, ++p)
        if (*p > 0)
            fprintf(stderr, " %+d(%d)", i, *p);
    fputc('\n', stderr);
}

int main(int argc, char *argv[])
{
    FILE *qf;
    const int max_buffer = 255;
    char buffer[max_buffer];
    struct sigaction sa;
    char *filename;
    char *ev;
    unsigned int random_seed;
    int l, count;
        
    sa.sa_handler = SIG_DFL;
    sa.sa_flags = SA_NOCLDSTOP;
    
    if (sigaction(SIGCHLD, &sa, NULL)) {
        perror(NULL);
        return EXIT_FAILURE;
    }

    fontsize = fontsize_str(12);

    if (argc < 2 || !argv[1][0]) {
out:
        fprintf(stderr, VERSION_AND_COPYRIGHT);
        fprintf(stderr, "Usage: %s <question file>[%s] [<basename>]\n", 
		argv[0], extension);
        fprintf(stderr, "(Hint: try using gentest rather than mktest!)\n");
        return EXIT_FAILURE;
    }

    if (argv[1][0] == '-' && argv[1][1] == 's' && argv[1][2] == '\0') {
	    if (argc < 3 || !argv[2][0]) 
		    goto out;
	    argv[1] = argv[2];
            fprintf(stderr, "%s: option '-s' is deprecated, "
                        "please use \\fontpitch{10} in ATQ file\n",
                         argv[0]);
	    fontsize = fontsize_str(10);
    }
		

    basename = NULL;

    if (argc > 2 && argv[2][0]) {
        basename = strdup(argv[2]);
        if (!basename)
            goto oom;
        l = strlen(basename);
        if (basename[l-1] == extension[3] && basename[l-2] == extension[2]
            && basename[l-3] == extension[1] && basename[l-4] == extension[0])
            basename[l-4] = '\0';
    }

    l = strlen(argv[1]);
    filename = argv[1];

    qf = fopen(filename, "r");

    if (!qf && (
        filename[l-1] != extension[3] || filename[l-2] != extension[2] 
        || filename[l-3] != extension[1] || filename[l-4] != extension[0])) {
        filename = (char *) malloc(sizeof(char)*(strlen(argv[1])+5));
        if (!filename) 
            goto oom;
        strcpy(filename, argv[1]);
        strcat(filename, extension);
        if (!basename)
            basename = argv[1];
    }
    
    if (!basename) {
        basename = strdup(argv[1]);
        if (!basename) {
oom:
            fprintf(stderr, "Memory allocation error\n");
            return EXIT_FAILURE;
        }
        basename[strlen(basename)-4]='\0';
    }            

    ev = strrchr(basename, '/');
    if (ev)
        basename = ev+1;

    if (!qf)
        qf = fopen(filename, "r");
    if (!qf) {
        perror(filename);
        return EXIT_FAILURE;
    }
   
    random_seed = time(NULL);
    srandom(random_seed);    

    count = readfile(qf);

    atd_header(count);
    
    fprintf(stderr, "Generating sheet #    ");

    while (!feof(stdin)) 
        if (fgets(buffer, max_buffer, stdin))
            process_line(buffer);
    
    show_form_stats();

    fprintf(stderr, "\rDone.                     \n");
    
    if (fclose(qf))
        perror(argv[1]);
    
    return EXIT_SUCCESS;
}



void process_line(char const *p)
{
    int form = 0, nquiz = 0;
    int quiz[MAXQUIZ];
    
#if 1
    if (p[strlen(p)-1] != '\n') {
        fprintf(stderr, "\nmktest: Broken line passed to process_line()\n");
        fprintf(stderr, "mktest: Offending line is \"%s\"\n", p);
        exit(EXIT_FAILURE);
    }
#endif
    
    while (*p && (*p == ' ' || *p == '\n'))
        ++p;
        
    if (!*p)
        return;

    while (*p && *p != ':') {
        if (*p < '0' || *p > '9') {
            fprintf(stderr, "Invalid number in standard input\n");
            exit(EXIT_FAILURE);
        }
        form = form * 10 + (*p) - '0';
        ++p;
    }
    
    ++p;
    
    fprintf(stderr, "\b\b\b\b%3d ", form+1);
    
    do {
        int tot = 0;
        while (*p && (*p == ' ' || *p == '\n'))
            ++p;
        while (*p && (*p != ' ' && *p != '\n')) {
            if (*p < '0' && *p > '9') {
                fprintf(stderr, "Invalid number in standard input\n");
                exit(EXIT_FAILURE);            
            }
            tot = tot * 10 + (*p) - '0';
            ++p;
        }
        if (*p) {
            quiz[nquiz++] = tot;
        }
    } while (*p);

    update_form_stats(quiz, nquiz);
    scramble(quiz, nquiz);    
    create_form(form, quiz, nquiz); 
    create_text(form, quiz, nquiz);
}

static void build_a_option(char s[], int quiz[], int n)
{
    int i;
    char *p = s;

    *p++ = '-';
    *p++ = 'a';
    for (i=0; i<n; ++i) {
        int c = nanswers[quiz[i]];
        if (c < 2 || c > MAXANSWERSPERQUESTION) {
            fprintf(stderr, "Invalid number of answers (%d) in question %d\n", c, quiz[i]);
            exit(EXIT_FAILURE);
        }
        *p++ = '0'+c;
        *p++ = ',';
    }
    *--p = '\0';
}

static void create_form(int form, int quiz[], int nquiz)
{
    const int str_buf=255;
    const int num_buf =16;

    char frm[num_buf], file[str_buf];
    char cour[str_buf], dte[str_buf];
    char fver[num_buf], ans[str_buf];

    char *ev, *argv[8];
    int rv, i=0;
    
    ev = getenv("MKFORMCMD");
    
    if (!ev || !*ev)
        ev = formprogram;
    argv[i++] = ev;
    
    if (snprintf(frm, num_buf, "%d", form) >= num_buf ||
        snprintf(file, str_buf, "%s-form-%03d.ps", basename, form) >= str_buf) {
over:
        fprintf(stderr, "Buffer overflow in create_form()\n");
        exit(EXIT_FAILURE);
    }
    
    if (snprintf(fver, num_buf, "-f%d", formversion) >= num_buf)
        goto over;
    argv[i++] = fver;

    build_a_option(ans, quiz, nquiz);
    argv[i++] = ans;

    if (formcourse && *formcourse) {
        if (snprintf(cour, str_buf, "-c%s", formcourse) >= str_buf)
            goto over;
        argv[i++] = cour;
    }

    if (formdate && *formdate) {
        if (snprintf(dte, str_buf, "-d%s", formdate) >= str_buf)
            goto over;
        argv[i++] = dte;
    }

    argv[i++] = frm;
    argv[i] = NULL;
    rv = execute_command(ev, argv, file, NULL);
    if (rv) {
        fprintf(stderr, "%s returned error code %d\nArguments: ", ev, rv);
        for(i=0;argv[i];++i)
            fprintf(stderr, "[%s] ", argv[i]);
        fprintf(stderr, "\n");
        exit(EXIT_FAILURE);
    }
}

void create_text(int form, int quiz[], int nquiz)
{
    int rv, i = 0;
    const int max_buffer = 255;
    char output[max_buffer], buffer[max_buffer];
    char *ev, *argv[8];
    FILE *wf;
    char *checkfile = "latexcmp.chk";
    char *rmcmd = "rm", *rmopt = "-f";
        
    printf("%03d:", form);
            
    if (snprintf(output, max_buffer, "%s-text-%03d.tex", basename, form) >= max_buffer) {
buffer_overflow:
        fprintf(stderr, "Buffer overflow in create_text()\n");
        exit(EXIT_FAILURE);
    }

    wf = fopen(output, "w");
    if (!wf) {
        perror(output);
        exit(EXIT_FAILURE);
    }
    
    
    fprintf(wf, TEXHEAD, fontsize, course, teacher, date, form,
            course, teacher, date, form, (preamble ?: DEFAULT_PREAMBLE), course, teacher,
            date, form, advise);

    for (i=0; i<nquiz; ++i) {
        int ans[MAXANSWERSPERQUESTION] = { 0, };
        int q=quiz[i];
        int count = 0;
        
        fprintf(wf, "\\question{%s%%\n}\\par\\bigskip\n", question[q]);
        printf(" %d%c", q+1, (mulquestion[q] ? '[' : '('));
            
        while (count < nanswers[q]) {
            int j = (int) ((float)(nanswers[q])*random()/(RAND_MAX+1.0));
            if (ans[j])
                continue;
            if (answer[q*MAXANSWERSPERQUESTION+j]) {
                fprintf(wf, "\\answer{%c}{%s%%\n}\\answerend\n", 'A'+count, 
                        answer[q*MAXANSWERSPERQUESTION+j]);
                printf("%c", j+'A');
                ++ans[j];
                ++count;
            } else {
                fprintf(stderr, "Null answer %d in question %d\n", j, q);
                exit(EXIT_FAILURE);
            }
        }
        printf("%c", (mulquestion[q] ? ']' : ')'));
    }
    
    fprintf(wf, TEXTAIL, checkfile, output);
    fclose(wf);

    ev = getenv("LATEXCMD");
    
    if (!ev || !*ev)
        ev = "latex";

    argv[0] = ev;
    argv[1] = output;
    argv[2] = NULL;

    rv = execute_command(ev, argv, "/dev/null", "/dev/null");
    
    if (rv == -1) 
        rv = error_in_latex_compilation(output, checkfile);
    
    if (rv > 0) {
        fprintf(stderr, "\nError in ATQ file, latex compilation failed.\n"
                        "Look into the %s file to find the error\n", output);
        argv[0] = rmcmd;
        argv[1] = rmopt;
        argv[2] = checkfile;
        argv[3] = NULL;
        execute_command(argv[0], argv, NULL, NULL);
        exit(EXIT_FAILURE);
    }
    
    output[strlen(output)-4] = '\0';
    
    ev = getenv("DVIPSCMD");
    
    if (!ev || !*ev)
        ev = "dvips";
    
    if (snprintf(buffer, max_buffer, "%s.ps", output) >= max_buffer)
        goto buffer_overflow;

    argv[0] = ev;
    argv[1] = "-o";
    argv[2] = buffer;
    argv[3] = output;
    argv[4] = NULL;
    
    execute_command(ev, argv, "/dev/null", "/dev/null");
         
    argv[0] = rmcmd;
    argv[1] = rmopt;
    argv[2] = buffer;
    rv = max_buffer;
    if (snprintf(buffer, rv, "%s.tex", output) >= rv)
        goto buffer_overflow;
    rv -= strlen(buffer)+1;
    argv[3] = buffer + strlen(buffer)+1;
    if (snprintf(argv[3], rv, "%s.log", output) >= rv)
        goto buffer_overflow;
    rv -= strlen(argv[3])+1;
    argv[4] = argv[3]+strlen(argv[3])+1;
    if (snprintf(argv[4], rv, "%s.dvi", output) >= rv)
        goto buffer_overflow;
    rv -= strlen(argv[4])+1;
    argv[5] = argv[4]+strlen(argv[4])+1;
    if (snprintf(argv[5], rv, "%s.aux", output) >= rv)
        goto buffer_overflow;
    argv[6] = checkfile;
    argv[7] = NULL;

    execute_command(argv[0], argv, NULL, NULL);
         
    printf("\n");        
}

void scramble(int *vect, int num)
{

#define _ITER 3

    int k, i, j, t;
    
    for (k=0; k<_ITER; ++k) {
        for (i=0; i<num; ++i) {
            j = (float) num * random() / (RAND_MAX + 1.0);
            
            /* ASSERT */
            if (j<0 || j>=num) {
                fprintf(stderr, "Index out of range in scramble()\n");
                exit(EXIT_FAILURE);
            }
            
            t = vect[j];
            vect[j] = vect[i];
            vect[i] = t;
        }
    }
    
#undef _ITER
}
